import string, regex, sys, regsub, socket
import data, mud, app

# parse command intended for client, (i.e. one prefaced by '#')
# allow abbreviations for most commands
def DoCommand(input):
    if not input: return
    mud.log('DoCommand: ' + input)
    words = string.split(input)

    # execute some python code?
    if input[0] == '@':
	input = regsub.sub("^@", '', input)
	exec input in globals()
	return

    #define/query an action?
    if string.find("action", words[0]) == 0: 
	Action(input)
	return

    #define/query an alias?
    if string.find("alias", words[0]) == 0: 
	Alias(words)
	return

    # clear session?
    if string.find("clear", words[0]) == 0: 
	Clear()
	return

    # send a carriage return?
    if string.find("cr", words[0]) == 0: 
	CR()
	return

    # resize the databuffer?
    if string.find("databuffer", words[0]) == 0: 
	DataBuffer(words)
	return
    
    # grep the databuffer?
    if string.find("datagrep", words[0]) == 0: 
	DataGrep(words)
	return
    
    # linegrep the databuffer?
    if string.find("datagreplines", words[0]) == 0: 
	DataGrepLines(words)
	return
    
    # gag some poor fool?
    if string.find("gag", words[0]) == 0: 
	Gag(words)
	return
    
    # do the history thing?
    if string.find("history", words[0]) == 0: 
	History(words)
	return
    
    # clear all sessions?
    if string.find("killall", words[0]) == 0: 
	KillAll()
	return

    # start logging?
    if string.find("log", words[0]) == 0: 
	Log(words)
	return

    # quit?
    if regex.search("^quit", input) != -1:
	Quit()

    # load aliases?
    if string.find("read", words[0]) == 0: 
	ParseFile(words)
	return

    # connect to a mud?
    if string.find("session", words[0]) == 0: 
	Ses(words[1:])
	return 'ses'

    # display something?
    if string.find("showme", words[0]) == 0: 
	Showme(words)
	return

    # define a substitute?
    if string.find("substitute", words[0]) == 0: 
	Substitute(words)
	return

    # toggle speedwalk?
    if string.find("speedwalk", words[0]) == 0: 
	SpeedWalk()
	return

    # delete an action?
    if string.find('unaction', words[0]) == 0:
	UnAction(words)
	return

    # send text from a file?
    if string.find('textin', words[0]) == 0:
	Textin(words)
	return

    # delete an alias?
    if string.find('unalias', words[0]) == 0:
	UnAlias(words)
	return

    # ungag said poor fool?
    if string.find("ungag", words[0]) == 0: 
	UnGag(words)
	return

    # remove a substitute?
    if string.find("unsubstitute", words[0]) == 0: 
	UnSubstitute(words)
	return

    # define/query variables?
    if string.find("variable", words[0]) == 0: 
	Variable(words)
	return

    # save aliases?
    if string.find("write", words[0]) == 0: 
	WriteFile(words)
	return

    # unrecognized command
    Putline("That command is not defined")
    return

# display something
def Showme(words):
    if len(words) == 1:
	Putline('showme what?')
	return
    PutUntouchedLine(string.join(words[1:]))

# Set active session
def SetSes(ses):
    data.currsession = ses
    ans = 'ok, session "' + ses.name + '" activated.'
    Putline(ans)

# Create a new session Connected to a mud.
# increment global session count and add new session to global
# session list
def Ses(to):
    if len(to) == 3:
	# see if there's an existing session with the same name
	for ses in data.sessionlist:
	    if ses.name == to[0]:
		Putline('session "'+ses.name+'" already exists.')		
		return
	try:

	    name = to[0]
	    host = to[1]
	    port = string.atoi(to[2])
	    Putline("Trying to connect...")
	    thisses = data.session(name,host,port)
	    
	except socket.error:
	    Putline("Unable to conect!")
	except ValueError:
	    Putline('illegal port number')
	else: # it worked
	    thisses.InitLocalSession()
	    data.currsession = thisses
	    data.sessionlist = data.sessionlist + [thisses]
	    data.numsessions = data.numsessions + 1

    elif len(to) == 2:
	Putline("Hey, specify a port number, willya?")
    elif len(to) == 1:
	Putline("Without an address, you can't mud.")
    else:
	Putline("usage: #session sesname host port")

# toggle speedwalking for the current session
def SpeedWalk():
    data.currsession.speedwalk = not data.currsession.speedwalk
    if data.currsession.speedwalk:
	Putline('speedwalking is now on')
    else:
	Putline('speedwalking is now off')

# the #databuffer command
# with one argument, set the size of currsession's databuffer
# with no arguments, display it
def DataBuffer(words):
    if len(words) < 2:
	Putline('databuffer size is %d'%data.currsession.databuf.size)
	return
    try:
	num = string.atoi(words[1])
    except ValueError:
	Putline('invalid argument')
    else:
	data.currsession.databuf.resize(num)
	Putline('databuffer size set to %d'%num)
	
# the #datagrep command
# search through the databuffer for a regex, printing all 
# matches in their entirety
def DataGrep(words):
    if len(words) < 2:
	Putline('command requires one argument')
	return
    pat = string.join(words[1:])
    got = data.currsession.databuf.grep(pat)
    for g in got:
	PutReallyUntouchedLine(g)
    Putline('%d match(es) found.'%len(got))

# the #datagreplines command
# search through the databuffer for a regex, printing 
# only the _lines_ which contain a match
def DataGrepLines(words):
    if len(words) < 2:
	Putline('command requires (at least) one argument')
	return
    pat = string.join(words[1:])
    build = data.currsession.databuf.greplines(pat)
    for b in build:
	PutReallyUntouchedLine(b)
    Putline('%d match(es) found.'%len(build))
    # just in case someone needs a return value someday
    return build

# the #variable command
# same argument scheme as aliases
def Variable(words):
    if len(words) == 1:
	# display all variables
	if len(data.currsession.vars.keys()) == 0:
	    Putline('you haven\'t even defined any variables!')
	else:
	    Putline('defined variables:')
	    for var in data.currsession.vars.keys():
		DisplayVar(var)
	return
    elif len(words) == 2:
	# display just the matching variables
	which = words[1]
	whichl = data.currsession.Expand(which, data.currsession.vars.keys())
	if not len(whichl):
	    Putline('that variable is not defined')
	else:
	    for w in whichl:
		DisplayVar(w)
    else:
	# more than one argument: define a new variable for the current session
	name, expansion = app.SplitBraced(string.join(words[1:]))
	if name and (not expansion):
	    # display just the matching variables
	    Variable(['#var', name])
	    return
	data.currsession.vars[name] = expansion
	Putline('variable defined:')
	DisplayVar(name)

# print a variable for the user
def DisplayVar(name):
    val = data.currsession.vars[name]
    ans = '{%s} = {%s}'%(name, val)
    PutUntouchedLine(ans)

# write aliases/actions, etc to a file
# this saves the local session and the global session in one fell swoop
def WriteFile(words):
    if len(words) != 2:
	Putline("command requires one argument")
	return
    al_count = 0
    ac_count = 0
    filename = data.initdir + words[1]
    thefile = open(filename, 'w')
    for ses in [data.currsession]:
	# aliases
	for al in ses.aliases.keys():
	    str = '#al {%s} {%s}\n'%(al, ses.aliases[al])
	    thefile.write(str)
	# actions
	for ac in ses.actions.keys():
	    str = '#ac {%s} {%s}\n'%(ac, ses.actions[ac])
	    thefile.write(str)	    
	# substitutes
	for sub in ses.subs.keys():
	    str = '#sub {%s} {%s}\n'%(sub, ses.subs[sub])
	    thefile.write(str)
	# gags
	for gag in ses.gags:
	    thefile.write('#gag ' + gag)
	# variables
	for var in ses.vars.keys():
	    str = '#var {%s} {%s}\n'%(var, ses.vars[var])
	    thefile.write(str)

    Putline('ok, session saved')

# the #textin command
# send text to the mud from a file
def Textin(words):
    if len(words) != 2:
	Putline('command requires one argument')
	return

    # if they give us a full path name, we try to open it.
    # otherwise we prepend the initdir to the argument
    if words[1][0] == '/':
	filename = words[1]
    else:
	filename = data.initdir + words[1]

    try:
	f = open(filename, 'r')
    except IOError:
	Putline('unable to open text file: ' + filename)
    else:
	# ok, got it open.  do the textin stuff...
	Putline('ok, sending commands...')
	thelist = f.readlines()
	for line in thelist:
	    data.theapp.HandleUserInput(line)

# read in aliases/actions/gags/substitutes from a file
def ParseFile(ofile):
    try:

	if len(ofile) != 2:
	    Putline("command requires one argument")
	    return

	# if they give us a full path name, we try to open it.
	# otherwise we prepend the initdir to the argument
	if ofile[1][0] == '/':
	    filename = ofile[1]
	else:
	    filename = data.initdir + ofile[1]
	thefile = open(filename, 'r')
	thelist = thefile.readlines()
	al_count = 0
	ac_count = 0
	sub_count = 0
	gag_count = 0
	var_count = 0

	# go through the file, adding actions, aliases etc where appropriate
	for s in thelist:
	    words = string.split(s)
	    if len(words) > 2:
		# alias
		if words[0] == '#al':
		    al_count = al_count + 1
		    name, expansion = app.SplitBraced(string.join(words[1:]))
		    data.currsession.aliases[name] = expansion
		# action
		elif words[0] == '#ac':
		    ac_count = ac_count + 1
		    trigger, response = SplitAction(s)
		    data.currsession.actions[trigger] = response
		# substitute
		elif string.find('#substitute', words[0]) == 0:
		    sub_count = sub_count + 1
		    pat, repl = app.SplitBraced(string.join(words[1]))
		    data.currsession.subs[pat] = repl
		# gag
		elif string.find('#gag', words[0]) == 0:
		    if len(words) > 1:
			gag_count = gag_count + 1
			data.currsession.gags = data.currsession.gags + \
						[string.join(words[1:])]
		#variable
		elif string.find('#variable', words[0]) == 0:
		    if len(words) > 2:
			var_count = var_count + 1
			name, val = app.SplitBraced(string.join(words[1:]))
			data.currsession.vars[name] = val


	Putline('ok.')
	Putline(string.join([str(al_count), "aliases loaded."]))
	Putline(string.join([str(ac_count), "actions loaded."]))
	Putline(string.join([str(sub_count), "substitutes loaded."]))
	Putline(string.join([str(var_count), "variables loaded."]))
	Putline(string.join([str(gag_count), "gags loaded."]))
	    
    except IOError: 
	Putline(string.join(["unable to open input file:", filename]))

# send a carriage return from the current session to its connection
# useful for aliases that want to send carriage returns
def CR():
    if data.currsession.connected:
	data.currsession.WriteTo('\n')

# start a log file for the current session
def Log(words):
    if len(words) == 1:
	# cancel an in-progress log
	if data.currsession.logging:
	    Putline('ok, closing logfile '+data.currsession.logfile.name)
	    data.currsession.logging = 0
	    data.currsession.logfile = None
	    return
	else:
	    # they aren't logging already, so they must have screwed up
	    Putline('log what?')
	    return
    if len(words) > 2:
	Putline('too many arguments')
	return
    if not data.currsession.connected:
	Putline('you aren\'t even connected!')
	return
    # if they give us a full path name, we try to open it.
    # otherwise we prepend the initdir to the argument
    if words[1][0] == '/':
	fullfile = words[1]
    else:
	fullfile = data.initdir + words[1]

    try:
	f = open(fullfile, 'w')
    except IOError:
	Putline('unable to open log file: ' + fullfile)
    else:
	# ok, got it open.  set up the logging stuff...
	Putline('ok, logging...')
	data.currsession.logging = 1
	data.currsession.logfile = f

# Quit lyntin
def Quit():
    Putline("bye...")
    if data.numsessions:
	data.ClearAll()
    sys.exit(0)

# wipe clean all active sessions.
# remove aliases/actions/subs/gags
def KillAll():
    for ses in data.sessionlist:
	ses.Clear()
	Putline('session "'+ses.name+'" cleared.')

# search input for speedwalking/aliases
# return a 2 elt tuple, of whether an alias was found,
# and what it expanded to
def WorkOver(input):
    ret = input
    splitup = string.split(input)
    whether = 0

    if splitup:	
	for ses in [data.currsession]: 
	    for each in ses.aliases.keys():
		if splitup[0] == each:
		    # found a match
		    whether = 1
		    # make a copy
		    ret = ses.aliases[each][:]
		    # get the variables that we'll need to fill in
		    varlist, ret = app.StripVars(ret)
		    varcount = len(varlist)
		    if varcount == 0:
			# this alias has no variables. just concatenate its expansion
			# with anything else the user typed, and return
			if len(splitup) > 1:
			    ret = ret + ' ' + string.join(splitup[1:])
			return whether, ret
		    
		    # find the values of the %0..%n variables
		    # %0 is the entire text (except for the command word of course)
		    # %n is the nth word after the command word
		    # unlike tintin, we don't restrict mudders to a mere nine variables
		    vars = {}
		    for v in varlist:
			n = string.atoi(v)
			if n == 0:
			    if len(splitup) > 1:
				vars[v] = string.join(splitup[1:])
			    else:
				vars[v] = ''
			else:
			    if len(splitup) > n:
				vars[v] = splitup[n]
			    else:
				vars[v] = ''

		    # go through the alias definition, substituting whatever
		    # the player typed for the variables %1...%n
		    for var in vars.keys():
			pat = '%' + var
			ret = regsub.gsub(pat, vars[var], ret)

    # check for speedwalk
    if regex.search('^[udnswe1-9][udnswe1-9]+$', input) != -1:
	if not whether:
	    whether = 'speed'
	    ret = input

    return whether, ret

# print a prompt
def Prompt():
    PutReallyUntouchedLine("\n>")

# return (trigger, response) from action string 
# input looks like '#ac {you see a dog} {kill dog}'
def SplitAction(ac): 
    ac_trigger = ''
    ac_response = ''
    
    # regex to match an action command
    action_regex = regex.compile(string.split(ac)[0] + ' {\(.*\)} {\(.*\)}')
    # I wrote this part before I knew how to do backreferences in python
    if action_regex.match(ac) != -1:
	ac_trigger = ac[action_regex.regs[1][0]:action_regex.regs[1][1]]
	ac_response = ac[action_regex.regs[2][0]:action_regex.regs[2][1]]
	return (ac_trigger, ac_response)

    # not defining an action, must be querying
    temp = string.split(ac)
    if len(temp) > 1:
	ac_trigger = string.join(temp[1:])

    # strip '{' and '}'
    if len(ac_trigger):
	if ac_trigger[0] == '{' and ac_trigger[-1] == '}':
	    ac_trigger = ac_trigger[1:-1]

    return (ac_trigger, ac_response)

# the #action command
# with no args, print all actions
# with one arg, print matching action(s)
# otherwise, create an action named arg1 which expands to
# the rest of the args
def Action(input):
    count = 0
    trigger, response = SplitAction(input) 
    
    if trigger and response:
	data.currsession.actions[trigger] = response
	PutUntouchedLine('#OK, {%s} NOW TRIGGERS {%s}'%(trigger, response))
	return

    elif trigger: 
	# print action definition
	for ses in [data.currsession]:
	    expanded = ses.ExpandAction(trigger)
	    if expanded:
		count = count + len(expanded)
		for ac in expanded:
		    PutUntouchedLine(string.join(['#ac '+'{'+ac+'}', "=",
						  '{'+ses.actions[ac]+'}']))

	if not count:
	    Putline("That action is not defined")
	return

    else: # print all current actions
	for ses in [data.currsession]:
	    for ac in ses.actions.keys():
		count = count + 1
		PutUntouchedLine(string.join(['#ac '+'{'+ac+'}', "=", 
					      '{'+ses.actions[ac]+'}']))
	if count == 0:
	    Putline("No actions defined.")
	return

# the #unaction command
# remove all matching actions
def UnAction(words):
    if len(words) < 2:
	Putline('command requires one argument')
	return

    un = string.join(words[1:])
    # remove braces if necessary
    if un[0] == '{' and un[-1] == '}':
	un = un[1:-1]
    acs = data.currsession.ExpandAction(un)
    if acs:
	for ac in acs:
	    # kill!!!
	    del data.currsession.actions[ac]
	Putline(str(len(acs)) + ' actions removed')
    else:
	Putline('that action is not defined')

# the #alias command
# with no args, print all aliases
# with one arg, print matching alias definition(s)
# with many args, define an alias named the first arg, which
# expands to the rest of the args
def Alias(words):
    count = 0
	
    if len(words) > 2:
	name, expansion = app.SplitBraced(string.join(words[1:]))
	data.currsession.aliases[name] = expansion
	PutUntouchedLine('#OK, {%s} ALIESES {%s}'%(name, expansion))
	return

    elif len(words) == 2: 
	# print alias definition
	name = words[1]
	for ses in [data.currsession]:
	    expanded = ses.ExpandAlias(name)
	    if expanded:
		count = count + len(expanded)
		for al in expanded:
		    PutUntouchedLine(string.join(['#al '+'{'+al+'}', " = ",
						  '{'+ses.aliases[al]+'}']))
	if not count:
	    Putline("That alias is not defined")
	return

    else: 
	# print all current aliases
	for ses in [data.currsession]:
	    for al in ses.aliases.keys():
		count = count + 1
		PutUntouchedLine(string.join(['#al '+'{'+al+'}', "=", 
					      '{'+ses.aliases[al]+'}']))
	if count == 0:
	    Putline("No aliases defined.")
	return

# the #history command
# with one numeric argument, set history size
# with no arguments, print last histsize commands
def History(words):
    if len(words) > 2:
	Putline('too many arguments')
	return
    elif len(words) == 2:
	# try to set a new history size
	try:
	    num = string.atoi(words[1])
	except ValueError:
	    Putline('invalid argument')
	    return
	else:
	    if num == 0:
		Putline('can\'t set history size to nothing.')
		return
	    data.histsize = num
	    Putline('ok, history size set to '+str(num))
    # print last histsize history entries
    else:
	n = len(data.history)
	if n == 0:
	    Putline('no history yet...')
	    return
	m = min([data.histsize, len(data.history)])
	for i in range(m - 1, -1, -1):
	    PutReallyUntouchedLine(str(i)+' '+str(data.history[i]))

# the #unalias command.
# remove all matching aliases
def UnAlias(words):
    if len(words) != 2:
	Putline('command requires one argument')
	return

    als = data.currsession.ExpandAlias(words[1])
    if als:
	for al in als:
	    # kill!!!
	    del data.currsession.aliases[al]
	Putline(str(len(als)) + ' aliases removed')	
    else:
	Putline('that alias is not defined')

# the #gag command.
# Cease displaying any text from the mud which contains
# the given string.
# useful for shutting up spammers
def Gag(words):
    if len(words) < 2:
	# print all gags
	if len(data.currsession.gags) == 0:
	    Putline('no gags are defined')
	else:
	    for gag in data.currsession.gags:
		Putline('gag ' + gag)
	return
    gagwhat = string.join(words[1:])
    # add string to current session's gags
    data.currsession.gags = data.currsession.gags + [gagwhat]
    Putline('ok, "' + gagwhat + '" is now gagged')

# the #ungag command.
# remove the given string from the session's gags
def UnGag(words):
    if len(words) < 2:
	Putline('command requires at least one argument')
	return
    ungagwhat = string.join(words[1:])
    ungagwhatlist = data.currsession.Expand(ungagwhat, data.currsession.gags)
    # remove ungagwhat from the current session's gags
    for g in ungagwhatlist:
	if data.currsession.gags.count(g) > 0:
	    data.currsession.gags.remove(g)
	    Putline('ok, "' + g + '" is no longer gagged')
    if not ungagwhatlist:
	Putline('that gag is not defined')


# the #substitute command.
# anytime we see a certain string from the mud, 
# substitute an alternate string for it.
def Substitute(words):
    if len(words) < 3:
	Putline('command requires at least two arguments')
	return
    pattern, replacement = app.SplitBraced(string.join(words[1:]))
    data.currsession.subs[pattern] = replacement
    Putline('ok, ' + pattern + ' is now replaced by ' + replacement)

# the #unsubstitute command
# remove substitute(s) from the current session
def UnSubstitute(words):
    if len(words) < 2:
	Putline('command requires at least one argument')
	return
    unlist = data.currsession.Expand(string.join(words[1:]), data.currsession.subs.keys())
    for sub in unlist:
	del data.currsession.subs[sub]
    Putline(str(len(unlist)) + ' substitutes removed')

# the #clear command
# Clear session of aliases, actions
def Clear():
    data.currsession.aliases = {}
    data.currsession.actions = {}
    Putline('session ' + data.currsession.name + ' cleared')


# print a message from the client to the player
# prepend a '#', and capitalize everything
def Putline(line):
    ret = string.upper(line)
    ret = "#" + ret
    PutUntouchedLine(ret)

def PutUntouchedLine(line):
    if line:
	print line

def PutReallyUntouchedLine(line):
    if line:
	print line, 
	#'**' + `len(line)` + '**',
